# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/transactionlist.py: ZFS transaction list

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Callable, Generator, List, Union

from tabulate import tabulate

from .abc import TransactionABC, TransactionListABC
from .debug import typechecked
from .i18n import t
from .io import colorize, humanize_size
from .transactionmeta import TransactionMetaTypes

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class TransactionList(TransactionListABC):
    """
    Mutable.
    """

    def __init__(self, *transactions: TransactionABC):

        self._transactions = list(transactions)
        self._changed = None

    def __len__(self) -> int:

        return len(self._transactions)

    def __add__(self, other: TransactionListABC) -> TransactionListABC:

        return type(self)(*(tuple(self.transactions) + tuple(other.transactions)))

    def __getitem__(self, index: int) -> TransactionABC:

        return self._transactions[index]

    @property
    def changed(self) -> Union[None, Callable]:

        return self._changed

    @changed.setter
    def changed(self, value: Union[None, Callable]):

        self._changed = value

    @property
    def table_columns(self) -> List[str]:

        headers = set()
        for transaction in self._transactions:
            keys = list(transaction.meta.keys())
            assert t("type") in keys
            headers.update(keys)
        headers = list(headers)
        headers.sort()

        if len(headers) == 0:
            return headers

        type_index = headers.index(t("type"))
        if type_index != 0:
            headers.pop(type_index)
            headers.insert(0, t("type"))

        return headers

    @property
    def table_rows(self) -> List[str]:

        return [f'{t("transaction"):s} #{index:d}' for index in range(1, len(self) + 1)]

    @property
    def transactions(self) -> Generator[TransactionABC, None, None]:

        return (transaction for transaction in self._transactions)

    def append(self, transaction: TransactionABC):

        self._transactions.append(transaction)
        if self._changed is not None:
            self._link_transaction(transaction)

    def extend(self, transactions: TransactionListABC):

        transactions = list(transactions.transactions)
        self._transactions.extend(transactions)
        if self._changed is not None:
            for transaction in transactions:
                self._link_transaction(transaction)

    def clear(self):

        self._transactions.clear()
        if self._changed is not None:
            self._changed()

    def _link_transaction(self, transaction: TransactionABC):

        transaction.changed = lambda: self._changed(
            self._transactions.index(transaction)
        )
        transaction.changed()

    def print_table(self):

        if len(self) == 0:
            return

        table_columns = self.table_columns
        colalign = self._table_colalign(table_columns)

        table = [
            [
                self._table_format_cell(header, transaction.meta.get(header))
                for header in table_columns
            ]
            for transaction in self._transactions
        ]

        print(
            tabulate(
                table,
                headers=table_columns,
                tablefmt="github",
                colalign=colalign,
            )
        )

    @staticmethod
    def _table_format_cell(
        header: str, value: Union[TransactionMetaTypes, None]
    ) -> str:

        FORMAT = {
            t("written"): lambda v: humanize_size(v, add_color=True),
        }

        return FORMAT.get(header, str)(value)

    @staticmethod
    def _table_colalign(headers: List[str]) -> List[str]:

        RIGHT = (t("written"),)
        DECIMAL = tuple()

        colalign = []
        for header in headers:
            if header in RIGHT:
                colalign.append("right")
            elif header in DECIMAL:
                colalign.append("decimal")
            else:
                colalign.append("left")

        return colalign

    def run(self):

        for transaction in self._transactions:

            print(
                f'({colorize(transaction.meta[t("type")], "white"):s}) '
                f'{colorize(str(transaction.command), "yellow"):s}'
            )

            assert not transaction.running
            assert not transaction.complete

            transaction.run()

            assert not transaction.running
            assert transaction.complete

            if transaction.error is not None:
                print(colorize(t("FAILED"), "red"))
                raise transaction.error
            else:
                print(colorize(t("OK"), "green"))



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/comparisonzpool.py: ZFS zpool comparison

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Generator, List, Union

from .abc import ComparisonItemABC, ComparisonZpoolABC, DatasetABC, ZpoolABC
from .comparisonitem import ComparisonItem
from .debug import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class ComparisonZpool(ComparisonZpoolABC):
    """
    Immutable. No order, just name-based matching.
    """

    def __init__(
        self,
        a: Union[ZpoolABC, None],
        b: Union[ZpoolABC, None],
        merged: List[ComparisonItemABC],
    ):

        assert a is not None or b is not None
        if a is not None and b is not None:
            assert type(a) == type(b)

        self._a, self._b, self._merged = a, b, merged

    def __len__(self) -> int:

        return len(self._merged)

    @property
    def a(self) -> Union[ZpoolABC, None]:

        return self._a

    @property
    def b(self) -> Union[ZpoolABC, None]:

        return self._b

    @property
    def merged(self) -> Generator[ComparisonItemABC, None, None]:

        return (item for item in self._merged)

    @staticmethod
    def _merge_datasets(
        items_a: Generator[DatasetABC, None, None],
        items_b: Generator[DatasetABC, None, None],
    ) -> List[ComparisonItemABC]:

        items_a = {item.subname: item for item in items_a}
        items_b = {item.subname: item for item in items_b}

        names = list(items_a.keys() | items_b.keys())
        merged = [
            ComparisonItem(items_a.get(name, None), items_b.get(name, None))
            for name in names
        ]
        merged.sort(key=lambda item: item.get_item().name)

        return merged

    @classmethod
    def from_zpools(
        cls,
        zpool_a: Union[ZpoolABC, None],
        zpool_b: Union[ZpoolABC, None],
    ) -> ComparisonZpoolABC:

        assert zpool_a is not None or zpool_b is not None

        if (zpool_a is None) ^ (zpool_b is None):
            return cls(
                a=zpool_a,
                b=zpool_b,
                merged=ComparisonItem.list_from_singles(
                    getattr(zpool_a, "datasets", None),
                    getattr(zpool_b, "datasets", None),
                ),
            )

        assert zpool_a is not zpool_b
        assert zpool_a != zpool_b

        return cls(
            a=zpool_a,
            b=zpool_b,
            merged=cls._merge_datasets(zpool_a.datasets, zpool_b.datasets),
        )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

	src/abgleich/__init__.py: Package root

	Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# META
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

__version__ = "0.1.1"


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# EXPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from .core.command import Command
from .core.config import Config
from .core.configfield import ConfigField
from .core.configspec import CONFIGSPEC
from .core.dataset import Dataset
from .core.debug import DEBUG
from .core.i18n import t
from .core.property import Property
from .core.snapshot import Snapshot
from .core.transaction import Transaction
from .core.transactionlist import TransactionList
from .core.transactionmeta import TransactionMeta
from .core.zpool import Zpool



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/configspec.py: Defines configuration fields

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from .configfield import ConfigField
from .i18n import t
from .lib import valid_name

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SPEC
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

CONFIGSPEC = [
    ConfigField(
        name="keep_snapshots",
        description=t(
            f"Defines how many snapshots should always be kept on the source side. Minimum is one."
        ),
        validate=lambda v: v >= 1,
        type_=int,
        default=1,
    ),
    ConfigField(
        name="keep_backlog",
        description=t(
            f"Degines how many snapshots should always be kept on the target side beyond the required overlap with the source side. Set to -1 for all."
        ),
        validate=lambda v: v >= -1,
        type_=int,
        default=-1,
    ),
    ConfigField(
        name="suffix",
        description=t(f"The suffix that snapshot names should carry."),
        validate=lambda v: valid_name(v, min_len=0),
        type_=str,
        default="",
    ),
    ConfigField(
        name="digits",
        description=t(
            f"Snapshots made on the same day get enumerated within their name. Defines how many digits this number should have."
        ),
        validate=lambda v: v >= 1,
        type_=int,
        default=2,
    ),
    ConfigField(
        name="always_changed",
        description=t(
            f"Assume that datasets and volumes have always been changed, forcing `abgleich snap` to create snapshots for all of them on every run."
        ),
        type_=bool,
        default=False,
    ),
    ConfigField(
        name="written_threshold",
        description=t(
            f"The creation of a snapshot is triggered if at least this many bytes have been written to a dataset or volume."
        ),
        validate=lambda v: v > 0,
        type_=int,
        default=1024**2,
    ),
    ConfigField(
        name="check_diff",
        description=t(
            f"Check the `diff` of a dataset if more than zero but less than `written_threshold` bytes have been written to it before creating a snapshot. Generating `diff`s can be slow or even fail under certain circumstances."
        ),
        type_=bool,
        default=True,
    ),
    ConfigField(
        name="ignore",
        description=t(
            f"Ignore the following comma-separated list of datasets and volumes for all operations."
        ),
        validate=lambda v: all((isinstance(item, str) and len(item) > 0 for item in v)),
        type_=list,
        import_=lambda v: list({item.strip() for item in v.split(',') if len(item.strip()) > 0}),
        default=list(),
    ),
    ConfigField(
        name="include_root",
        description=t(
            f"Include the root (source prefix & target prefix) into all operations."
        ),
        type_=bool,
        default=True,
    ),
    ConfigField(
        name="compatibility/tagging",
        description=t(
            f"Make `abgleich` work around snapshots created by other tools on the source side."
        ),
        type_=bool,
        default=False,
    ),
    ConfigField(
        name="compatibility/target_samba_noshare",
        description=t(
            f"Tell ZFS on the target side to not expose datasets and volumes via NFS."
        ),
        type_=bool,
        default=False,
    ),
    ConfigField(
        name="compatibility/target_autosnapshot_ignore",
        description=t(
            f"Tell `zfs-auto-snapshot` on the target side to ignore datasets and volumes."
        ),
        type_=bool,
        default=False,
    ),
    ConfigField(
        name="ssh/compression",
        description=t(f"Activate compression during `ssh` transfers."),
        type_=bool,
        default=False,
    ),
    ConfigField(
        name="ssh/cipher",
        description=t(f"Specify `OpenSSH` cipher."),
        type_=str,
        default="",
    ),
]

for _side in ("source", "target"):

    CONFIGSPEC.extend(
        [
            ConfigField(
                name=f"{_side}/zpool",
                description=t(f"Name of zpool on {_side:s} side."),
                validate=lambda v: len(v) > 0,
                type_=str,
            ),
            ConfigField(
                name=f"{_side}/prefix",
                description=t(f"Path to root dataset/volume in zpool on {_side:s} side."),
                type_=str,
                default="",
            ),
            ConfigField(
                name=f"{_side}/host",
                description=t(f"Hostname or IP address of {_side:s} side."),
                validate=lambda v: len(v) > 0,
                type_=str,
                default="localhost",
            ),
            ConfigField(
                name=f"{_side}/user",
                description=t(f"Username on {_side:s} side (SSH)."),
                type_=str,
                default="",
            ),
            ConfigField(
                name=f"{_side}/port",
                description=t(f'Port on {_side:s} side (SSH). "0" indicates SSH to fall back to its own configuration.'),
                validate=lambda v: v >= 0,
                type_=int,
                default=0,
            ),
            ConfigField(
                name=f"{_side}/processing",
                description=t(
                    "Pre-processing command on source side." if _side == 'source' else "Post-processing command on target side."
                ),
                type_=str,
                default="",
            ),
        ]
    )

del _side



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

	tests/test_config.py: Configuration module tests

	Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONST
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

RAW = """always_changed: false
check_diff: false
digits: 2
ignore:
- CACHE
- CACHE_CC
- CACHE_THUMBNAILS
- DESKTOP/SCRATCH
include_root: true
keep_backlog: true
keep_snapshots: 2
source:
    host: sourcehost
    prefix: sourceprefix
    zpool: sourcepool
ssh:
    cipher: aes256-gcm@openssh.com
    compression: false
suffix: _backup
target:
    host: targethost
    prefix: targetprefix
    user: targetuser
    zpool: targetpool
written_threshold: 1048576
"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from deepdiff import DeepDiff

from abgleich import Config

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def test_importexport():

    c = Config.from_text(RAW)
    raw = c.to_text()

    assert RAW == raw

    d = Config.from_text(raw)

    assert DeepDiff(c, d, ignore_order=True) == {}



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/gui/transaction.py: ZFS transactions

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Callable, Union

from PyQt5.QtCore import QAbstractTableModel, QModelIndex, Qt
from PyQt5.QtGui import QColor

from ..core.abc import TransactionListABC
from ..core.debug import typechecked
from ..core.io import humanize_size
from ..core.i18n import t

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class TransactionListModel(QAbstractTableModel):
    """
    Mutable.
    """

    def __init__(self, transactions: TransactionListABC, parent_changed: Callable):

        super().__init__()
        self._transactions = transactions
        self._transactions.changed = self._transactions_changed
        self._parent_changed = parent_changed

        self._rows, self._cols = None, None
        self._update_labels()

    def data(
        self, index: QModelIndex, role: int
    ) -> Union[None, str, QColor]:  # TODO return type

        row, col = index.row(), index.column()
        col_key = self._cols[col]

        if role == Qt.DisplayRole:
            if col_key == t("written"):
                return humanize_size(self._transactions[row].meta[col_key])
            return self._transactions[row].meta[col_key]

        if role == Qt.ForegroundRole:
            if col_key != t("written"):
                return
            return QColor("#808080")

        if role == Qt.BackgroundRole:
            if col_key != t("written"):
                return
            return QColor(
                humanize_size(self._transactions[row].meta[col_key], get_rgb=True)
            )

        if role == Qt.DecorationRole:
            if col_key != t("type"):
                return
            if self._transactions[row].error is not None:
                return QColor("#FF0000")
            if self._transactions[row].running:
                return QColor("#FFFF00")
            if self._transactions[row].complete:
                return QColor("#00FF00")
            return QColor("#0000FF")

    def headerData(
        self, section: int, orientation: Qt.Orientation, role: int
    ) -> Union[None, str]:

        if role == Qt.DisplayRole:

            if orientation == Qt.Horizontal:
                return self._cols[section]

            if orientation == Qt.Vertical:
                return self._rows[section]

    def rowCount(self, index: QModelIndex) -> int:

        return len(self._transactions)

    def columnCount(self, index: QModelIndex) -> int:

        return len(self._cols)

    def _transactions_changed(self, row: Union[None, int] = None):

        old_rows, old_cols = self._rows, self._cols
        self._update_labels()
        if old_rows != self._rows or old_cols != self._cols:
            self.layoutChanged.emit()
        if row is not None:
            self.dataChanged.emit(
                self.index(row, 0), self.index(row, len(self._cols) - 1)
            )
        self._parent_changed()

    def _update_labels(self):

        self._rows = self._transactions.table_rows
        self._cols = self._transactions.table_columns



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/command.py: Sub-process wrapper for commands

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import itertools
from subprocess import Popen, PIPE
from typing import List, Optional, Tuple, Union
import shlex

from .abc import CommandABC, ConfigABC
from .debug import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Command(CommandABC):
    """
    Immutable.
    """

    def __init__(self, cmd: List[List[str]]):

        self._cmd = [fragment.copy() for fragment in cmd]

    def __repr__(self) -> str:

        return "<Command>"

    def __str__(self) -> str:

        return " | ".join([shlex.join(fragment) for fragment in self._cmd])

    def __len__(self) -> int:

        return len(self._cmd)

    def __or__(self, other: CommandABC) -> CommandABC:  # pipe

        return type(self)(self.cmd + other.cmd)

    @staticmethod
    def _com_to_str(
        com: Union[str, bytes, None], max_byte_chars: Optional[int] = None
    ) -> str:

        if com is None:
            return ""

        if not isinstance(com, bytes):
            return com

        try:
            com = com.decode("utf-8")
        except UnicodeDecodeError:
            com = repr(com)

        if max_byte_chars is not None:
            if len(com) > max_byte_chars:
                return f"[{len(com)-max_byte_chars:d} characters ...]\n{com[-max_byte_chars:]:s}"

        return com

    @staticmethod
    def _split_list(data: List, delimiter: str) -> List[List]:

        return [
            list(sub_list)
            for is_delimiter, sub_list in itertools.groupby(
                data, lambda item: item == delimiter
            )
            if not is_delimiter
        ]

    def run(
        self,
        returncode: bool = False,
        max_out_byte_chars: Optional[int] = None,
        max_err_byte_chars: Optional[int] = None,
    ) -> Union[
        Tuple[List[str], List[str], List[int], Exception], Tuple[List[str], List[str]]
    ]:

        procs = []  # all processes, connected with pipes

        for index, fragment in enumerate(self._cmd):  # create & connect processes

            stdin = None if index == 0 else procs[-1].stdout  # output of last process
            proc = Popen(
                fragment,
                stdout=PIPE,
                stderr=PIPE,
                stdin=stdin,
            )
            procs.append(proc)

        output, errors, status = [], [], []

        for proc in procs[::-1]:  # inverse order, last process first

            out, err = proc.communicate()
            output.append(self._com_to_str(out, max_byte_chars=max_out_byte_chars))
            errors.append(self._com_to_str(err, max_byte_chars=max_err_byte_chars))
            status.append(int(proc.returncode))

        output.reverse()
        errors.reverse()
        status.reverse()

        exception = SystemError("command failed", str(self), output, errors)

        if returncode:
            return output, errors, status, exception

        if any((code != 0 for code in status)):  # some fragment failed:
            raise exception

        return output, errors

    def on_side(self, side: str, config: ConfigABC) -> CommandABC:

        if config[f"{side:s}/host"].value == "localhost":
            return self

        side_config = config.group(side)
        ssh_config = config.group("ssh")

        cmd_ssh = [
            "ssh",
            "-T",  # Disable pseudo-terminal allocation
            "-o",  # Option parameter
            "Compression=yes" if ssh_config["compression"].value else "Compression=no",
        ]
        if side_config["port"].value != 0:
            cmd_ssh.extend(["-p", f'{side_config["port"].value:d}'])
        if ssh_config["cipher"].value is not None:
            cmd_ssh.extend(("-c", ssh_config["cipher"].value))
        cmd_ssh.extend(
            [f'{side_config["user"].value:s}@{side_config["host"].value:s}', str(self)]
        )

        return type(self)([cmd_ssh])

    @property
    def cmd(self) -> List[List[str]]:

        return [fragment.copy() for fragment in self._cmd]

    @classmethod
    def from_str(cls, cmd: str) -> CommandABC:

        return cls(cls._split_list(shlex.split(cmd), "|"))

    @classmethod
    def from_list(cls, cmd: List[str]) -> CommandABC:

        return cls([cmd])



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/transactionmeta.py: ZFS transaction meta

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Generator, Union

from .abc import TransactionMetaABC
from .debug import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TYPES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

TransactionMetaTypes = Union[str, int, float]

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class TransactionMeta(TransactionMetaABC):
    """
    Immutable.
    """

    def __init__(self, **kwargs: TransactionMetaTypes):

        self._meta = kwargs

    def __getitem__(self, key: str) -> TransactionMetaTypes:

        return self._meta[key]

    def __len__(self) -> int:

        return len(self._meta)

    def get(self, key: str) -> Union[TransactionMetaTypes, None]:

        return self._meta.get(key, None)

    def keys(self) -> Generator[str, None, None]:

        return (key for key in self._meta.keys())



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/zpool.py: ZFS zpool

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from collections import OrderedDict
from typing import Generator, List, Tuple, Union

from tabulate import tabulate

from .abc import (
    ComparisonItemABC,
    ConfigABC,
    DatasetABC,
    SnapshotABC,
    TransactionListABC,
    ZpoolABC,
)
from .command import Command
from .comparisondataset import ComparisonDataset
from .comparisonzpool import ComparisonZpool
from .dataset import Dataset
from .debug import typechecked
from .i18n import t
from .io import colorize, humanize_size
from .lib import join, root
from .property import Property
from .transaction import Transaction
from .transactionlist import TransactionList

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Zpool(ZpoolABC):
    """
    Immutable.
    """

    def __init__(
        self,
        datasets: List[DatasetABC],
        side: str,
        config: ConfigABC,
    ):

        self._datasets = datasets
        self._side = side
        self._config = config

        self._root = root(
            config[f"{side:s}/zpool"].value, config[f"{side:s}/prefix"].value
        )

    def __eq__(self, other: ZpoolABC) -> bool:

        return self.side == other.side

    @property
    def datasets(self) -> Generator[DatasetABC, None, None]:

        return (dataset for dataset in self._datasets)

    @property
    def side(self) -> str:

        return self._side

    @property
    def root(self) -> str:

        return self._root

    def get_cleanup_transactions(
        self,
        other: ZpoolABC,
    ) -> TransactionListABC:

        assert self.side != other.side
        assert self.side in ("source", "target")
        assert other.side in ("source", "target")

        zpool_comparison = ComparisonZpool.from_zpools(self, other)
        transactions = TransactionList()

        for dataset_item in zpool_comparison.merged:
            transactions.extend(self._get_cleanup_from_datasetitem(dataset_item))

        return transactions

    def generate_cleanup_transactions(
        self,
        other: ZpoolABC,
    ) -> Generator[Tuple[int, Union[None, TransactionListABC]], None, None]:

        assert self.side != other.side
        assert self.side in ("source", "target")
        assert other.side in ("source", "target")

        zpool_comparison = ComparisonZpool.from_zpools(self, other)

        yield len(zpool_comparison), None

        for index, dataset_item in enumerate(zpool_comparison.merged):
            yield index, self._get_cleanup_from_datasetitem(dataset_item)

    def _get_cleanup_from_datasetitem(
        self,
        dataset_item: ComparisonItemABC,
    ) -> TransactionListABC:

        if dataset_item.get_item().ignore:
            return TransactionList()
        if dataset_item.a is None or dataset_item.b is None:
            return TransactionList()
        if self.side == "target" and self._config["keep_backlog"].value == -1:
            return TransactionList()

        dataset_comparison = ComparisonDataset.from_datasets(
            dataset_item.a, dataset_item.b, self._config
        )  # TODO namespace

        if self.side == "source":
            snapshots = dataset_comparison.a_overlap_tail[
                : -self._config["keep_snapshots"].value
            ]
        else:  # target
            if self._config["keep_backlog"].value == 0:
                keep_backlog = None
            else:
                keep_backlog = -self._config["keep_backlog"].value
            snapshots = dataset_comparison.a_disjoint_tail[:keep_backlog]

        transactions = TransactionList()

        for snapshot in snapshots:
            transactions.extend(snapshot.get_cleanup_transactions())

        return transactions

    def get_backup_transactions(
        self,
        other: ZpoolABC,
    ) -> TransactionListABC:

        assert self.side == "source"
        assert other.side == "target"

        zpool_comparison = ComparisonZpool.from_zpools(self, other)
        transactions = TransactionList()

        for dataset_item in zpool_comparison.merged:
            transactions.extend(
                self._get_backup_transactions_from_datasetitem(other, dataset_item)
            )

        transactions.extend(self._get_backup_propery_transactions(other))

        return transactions

    def generate_backup_transactions(
        self,
        other: ZpoolABC,
    ) -> Generator[Tuple[int, Union[None, TransactionListABC]], None, None]:

        assert self.side == "source"
        assert other.side == "target"

        zpool_comparison = ComparisonZpool.from_zpools(self, other)

        yield len(zpool_comparison), None

        for index, dataset_item in enumerate(zpool_comparison.merged):
            yield index, self._get_backup_transactions_from_datasetitem(
                other, dataset_item
            )

        for transaction in self._get_backup_propery_transactions(other):
            yield len(zpool_comparison) - 1, transaction

    def _get_backup_transactions_from_datasetitem(
        self,
        other: ZpoolABC,
        dataset_item: ComparisonItemABC,
    ) -> TransactionListABC:

        if dataset_item.get_item().ignore:
            return TransactionList()
        if dataset_item.a is None:
            return TransactionList()

        if dataset_item.b is None:
            snapshots = list(dataset_item.a.snapshots)
        else:
            dataset_comparison = ComparisonDataset.from_datasets(
                dataset_item.a, dataset_item.b, self._config
            )  # TODO namespace
            snapshots = dataset_comparison.a_disjoint_head

        if len(snapshots) == 0:
            return TransactionList()

        source_dataset = (
            self.root
            if len(dataset_item.a.subname) == 0
            else join(self.root, dataset_item.a.subname)
        )
        target_dataset = (
            other.root
            if len(dataset_item.a.subname) == 0
            else join(other.root, dataset_item.a.subname)
        )

        transactions = TransactionList()

        for snapshot in snapshots:
            transactions.extend(
                snapshot.get_backup_transactions(
                    source_dataset,
                    target_dataset,
                )
            )

        return transactions

    def _get_backup_propery_transactions(self, other: ZpoolABC) -> TransactionListABC:

        transactions = TransactionList()

        if self._config["compatibility/target_samba_noshare"].value:
            transactions.append(
                Transaction.set_property(
                    item=other.root,
                    property=Property(name="sharesmb", value="off"),
                    side="target",
                    config=self._config,
                )
            )

        if self._config["compatibility/target_autosnapshot_ignore"].value:
            transactions.append(
                Transaction.set_property(
                    item=other.root,
                    property=Property(name="com.sun:auto-snapshot", value="false"),
                    side="target",
                    config=self._config,
                )
            )

        return transactions

    def get_snapshot_transactions(self) -> TransactionListABC:

        assert self._side == "source"

        transactions = TransactionList()

        for dataset in self._datasets:
            transactions.extend(self._get_snapshot_transactions_from_dataset(dataset))

        return transactions

    def generate_snapshot_transactions(
        self,
    ) -> Generator[Tuple[int, Union[None, TransactionListABC]], None, None]:

        assert self._side == "source"

        yield len(self._datasets), None

        for index, dataset in enumerate(self._datasets):
            yield index, self._get_snapshot_transactions_from_dataset(dataset)

    def _get_snapshot_transactions_from_dataset(
        self, dataset: DatasetABC
    ) -> TransactionListABC:

        if dataset.ignore:
            return TransactionList()
        if (
            dataset.get("mountpoint").value is None
            and dataset["type"].value == "filesystem"
        ):
            return TransactionList()
        if not dataset.changed:  # TODO namespace
            return TransactionList()

        return dataset.get_snapshot_transactions()

    def print_table(self):

        table = []
        for dataset in self._datasets:
            table.append(self._table_row(dataset, ignore=dataset.ignore))
            for snapshot in dataset.snapshots:
                table.append(self._table_row(snapshot))

        if len(table) == 0:
            print("(empty)")
            return

        print(
            tabulate(
                table,
                headers=(t("NAME"), t("USED"), t("REFER"), t("compressratio")),
                tablefmt="github",
                colalign=("left", "right", "right", "decimal"),
            )
        )

    @staticmethod
    def _table_row(
        entity: Union[SnapshotABC, DatasetABC], ignore: bool = False
    ) -> List[str]:

        color = "white" if not ignore else "red"

        return [
            "- " + colorize(entity.name, "grey")
            if isinstance(entity, SnapshotABC)
            else colorize(entity.name, color),
            humanize_size(entity["used"].value, add_color=True),
            humanize_size(entity["referenced"].value, add_color=True),
            f'{entity["compressratio"].value:.02f}',
        ]

    def print_comparison_table(self, other: ZpoolABC):

        zpool_comparison = ComparisonZpool.from_zpools(self, other)
        table = []

        for dataset_item in zpool_comparison.merged:
            table.append(
                self._comparison_table_row(
                    dataset_item, ignore=dataset_item.get_item().ignore
                )
            )
            if dataset_item.complete:
                dataset_comparison = ComparisonDataset.from_datasets(
                    dataset_item.a, dataset_item.b, self._config
                )
            elif dataset_item.a is not None:
                dataset_comparison = ComparisonDataset.from_datasets(
                    dataset_item.a, None, self._config
                )
            else:
                dataset_comparison = ComparisonDataset.from_datasets(
                    None, dataset_item.b, self._config
                )
            for snapshot_item in dataset_comparison.merged:
                table.append(
                    self._comparison_table_row(
                        snapshot_item, ignore=dataset_item.get_item().ignore
                    )
                )

        print(
            tabulate(
                table,
                headers=[t("NAME"), t(self.side), t(other.side)],
                tablefmt="github",
            )
        )

    @staticmethod
    def _comparison_table_row(
        item: ComparisonItemABC, ignore: bool = False
    ) -> List[str]:

        color1, color2 = ("white", "grey") if not ignore else ("red", "yellow")
        colorR, colorG, colorB = (
            ("red", "green", "blue") if not ignore else ("grey", "grey", "grey")
        )

        symbol = "X"

        entity = item.get_item()
        name = entity.name if isinstance(entity, SnapshotABC) else entity.subname

        if item.a is not None and item.b is not None:
            a, b = colorize(symbol, colorG), colorize(symbol, colorG)
        elif item.a is None and item.b is not None:
            a, b = "", colorize(symbol, colorB)
        elif item.a is not None and item.b is None:
            a, b = colorize(symbol, colorR), ""

        return [
            "- " + colorize(name, color2)
            if isinstance(entity, SnapshotABC)
            else colorize(name, color1),
            a,
            b,
        ]

    @staticmethod
    def available(
        side: str,
        config: ConfigABC,
    ) -> int:

        output, _ = (
            Command.from_list(
                [
                    "zfs",
                    "get",
                    "available",
                    "-H",
                    "-p",
                    root(
                        config[f"{side:s}/zpool"].value,
                        config[f"{side:s}/prefix"].value,
                    ),
                ]
            )
            .on_side(side=side, config=config)
            .run()
        )

        return Property.from_params(*output[0].strip().split("\t")[1:]).value

    @classmethod
    def from_config(
        cls,
        side: str,
        config: ConfigABC,
    ) -> ZpoolABC:

        root_dataset = root(
            config[f"{side:s}/zpool"].value, config[f"{side:s}/prefix"].value
        )

        output, errors, returncode, exception = (
            Command.from_list(
                [
                    "zfs",
                    "get",
                    "all",
                    "-r",
                    "-H",
                    "-p",
                    root_dataset,
                ]
            )
            .on_side(side=side, config=config)
            .run(returncode=True)
        )

        if returncode[0] != 0 and "dataset does not exist" in errors[0]:
            return cls(
                datasets=[],
                side=side,
                config=config,
            )
        if returncode[0] != 0:
            raise exception

        output = [
            line.split("\t") for line in output[0].split("\n") if len(line.strip()) > 0
        ]
        entities = OrderedDict((line[0], []) for line in output)
        for line_list in output:
            entities[line_list[0]].append(line_list[1:])

        if not config["include_root"].value:
            entities.pop(root_dataset)
            for name in [
                snapshot
                for snapshot in entities.keys()
                if snapshot.startswith(f"{root_dataset:s}@")
            ]:
                entities.pop(name)

        datasets = [
            Dataset.from_entities(
                name,
                OrderedDict(
                    (k, v)
                    for k, v in entities.items()
                    if k == name or k.startswith(f"{name:s}@")
                ),
                side,
                config,
            )
            for name in entities.keys()
            if "@" not in name
        ]
        datasets.sort(key=lambda dataset: dataset.name)

        return cls(
            datasets=datasets,
            side=side,
            config=config,
        )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    docs/source/conf.py: Configures the documentation build process

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

from docs.source.version import get_version

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "abgleich"
author = "Sebastian M. Ernst"
copyright = f"2019-2026 {author:s}"
release = get_version()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_click",
    "sphinx_rtd_theme",
    "myst_parser",
    "sphinx_autodoc_typehints",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_context = {
    "sidebar_external_links_caption": "Links",
    "sidebar_external_links": [
        # ('<i class="fa fa-rss fa-fw"></i> Blog', 'https://www.000'),
        (
            '<i class="fa fa-github fa-fw"></i> Source Code',
            "https://github.com/pleiszenburg/abgleich",
        ),
        (
            '<i class="fa fa-bug fa-fw"></i> Issue Tracker',
            "https://github.com/pleiszenburg/abgleich/issues",
        ),
        # ('<i class="fa fa-envelope fa-fw"></i> Mailing List', 'https://groups.io/g/abgleich-dev'),
        (
            '<i class="fa fa-comments fa-fw"></i> Chat',
            "https://matrix.to/#/#abgleich:matrix.org",
        ),
        # ('<i class="fa fa-file-text fa-fw"></i> Citation', 'https://doi.org/000'),
        (
            '<i class="fa fa-info-circle fa-fw"></i> pleiszenburg.de',
            "https://www.pleiszenburg.de/",
        ),
    ],
}

always_document_param_types = True  # sphinx_autodoc_typehints

napoleon_include_special_with_doc = True  # napoleon
# napoleon_use_param = True
# napoleon_type_aliases = True



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/transaction.py: ZFS transactions

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Callable, Union

from .abc import CommandABC, ConfigABC, PropertyABC, TransactionABC, TransactionMetaABC
from .command import Command
from .debug import typechecked
from .i18n import t
from .transactionmeta import TransactionMeta

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Transaction(TransactionABC):
    """
    Mutable.
    """

    def __init__(
        self,
        meta: TransactionMetaABC,
        command: CommandABC,
    ):

        self._meta, self._command = meta, command

        self._complete = False
        self._running = False
        self._error = None

        self._changed = None

    @property
    def changed(self) -> Union[None, Callable]:

        return self._changed

    @changed.setter
    def changed(self, value: Union[None, Callable]):

        self._changed = value

    @property
    def complete(self) -> bool:

        return self._complete

    @property
    def command(self) -> CommandABC:

        return self._command

    @property
    def error(self) -> Union[Exception, None]:

        return self._error

    @property
    def meta(self) -> TransactionMetaABC:

        return self._meta

    @property
    def running(self) -> bool:

        return self._running

    def run(self):

        if self._complete:
            return

        self._running = True
        if self._changed is not None:
            self._changed()

        try:
            _, _ = self._command.run(
                max_out_byte_chars=102_400,
                max_err_byte_chars=102_400,
            )
        except SystemError as error:
            self._error = error
        finally:
            self._running = False
            self._complete = True
            if self._changed is not None:
                self._changed()

    @classmethod
    def set_property(
        cls,
        item: str,
        property: PropertyABC,
        side: str,
        config: ConfigABC,
    ) -> TransactionABC:

        return cls(
            meta=TransactionMeta(
                **{
                    t("type"): t("set_property"),
                    t("item"): item,
                }
            ),
            command=Command.from_list(
                [
                    "zfs",
                    "set",
                    f"{property.name:s}={property.value_export:s}",
                    item,
                ]
            ).on_side(side=side, config=config),
        )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/abc.py: Abstract base classes

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import ABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASSES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


class CloneABC(ABC):
    pass


class CommandABC(ABC):
    pass


class ComparisonDatasetABC(ABC):
    pass


class ComparisonItemABC(ABC):
    pass


class ComparisonZpoolABC(ABC):
    pass


class ConfigABC(ABC):
    pass


class ConfigFieldABC(ABC):
    pass


class DatasetABC(ABC):
    pass


class PropertyABC(ABC):
    pass


class SnapshotABC(ABC):
    pass


class TransactionABC(ABC):
    pass


class TransactionListABC(ABC):
    pass


class TransactionMetaABC(ABC):
    pass


class ZpoolABC(ABC):
    pass



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/comparisondataset.py: ZFS dataset comparison

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import itertools
from typing import Generator, List, Union

from .abc import (
    ComparisonDatasetABC,
    ComparisonItemABC,
    ConfigABC,
    DatasetABC,
    SnapshotABC,
)
from .comparisonitem import ComparisonItem, ComparisonItemType, ComparisonStrictItemType
from .debug import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class ComparisonDataset(ComparisonDatasetABC):
    """
    Immutable.
    """

    def __init__(
        self,
        a: Union[DatasetABC, None],
        b: Union[DatasetABC, None],
        merged: List[ComparisonItemABC],
    ):

        assert a is not None or b is not None
        if a is not None and b is not None:
            assert type(a) == type(b)

        self._a, self._b, self._merged = a, b, merged

    def __len__(self) -> int:

        return len(self._merged)

    @property
    def a(self) -> Union[DatasetABC, None]:

        return self._a

    @property
    def a_disjoint_head(self) -> List[ComparisonStrictItemType]:

        return self._disjoint_head(
            source=[item.a for item in self._merged],
            target=[item.b for item in self._merged],
        )

    @property
    def a_disjoint_tail(self) -> List[ComparisonStrictItemType]:

        return self._disjoint_head(
            source=[item.a for item in self._merged][::-1],
            target=[item.b for item in self._merged][::-1],
        )[::-1]

    @property
    def a_overlap_tail(self) -> List[ComparisonStrictItemType]:

        return self._overlap_tail(
            source=[item.a for item in self._merged],
            target=[item.b for item in self._merged],
        )

    @property
    def b(self) -> Union[DatasetABC, None]:

        return self._b

    @property
    def b_disjoint_head(self) -> List[ComparisonStrictItemType]:

        return self._disjoint_head(
            source=[item.b for item in self._merged],
            target=[item.a for item in self._merged],
        )

    @property
    def b_disjoint_tail(self) -> List[ComparisonStrictItemType]:

        return self._disjoint_head(
            source=[item.b for item in self._merged][::-1],
            target=[item.a for item in self._merged][::-1],
        )[::-1]

    @property
    def b_overlap_tail(self) -> List[ComparisonStrictItemType]:

        return self._overlap_tail(
            source=[item.b for item in self._merged],
            target=[item.a for item in self._merged],
        )

    @property
    def merged(self) -> Generator[ComparisonItemABC, None, None]:

        return (item for item in self._merged)

    @classmethod
    def _disjoint_head(
        cls,
        source: List[ComparisonItemType],
        target: List[ComparisonItemType],
    ) -> List[ComparisonItemType]:
        """
        Returns new elements from source.
        If target is empty, returns source.
        If head of target and head of source are identical, returns empty list.
        """

        source, target = cls._strip_none(source), cls._strip_none(target)

        if any((element is None for element in source)):
            raise ValueError("source is not consecutive")
        if any((element is None for element in target)):
            raise ValueError("target is not consecutive")

        if len(source) == 0:
            raise ValueError("source must not be empty")

        if len(set([item.name for item in source])) != len(source):
            raise ValueError("source contains doublicate entires")
        if len(set([item.name for item in target])) != len(target):
            raise ValueError("target contains doublicate entires")

        if len(target) == 0:
            return source  # all of source, target is empty

        try:
            source_index = [item.name for item in source].index(target[-1].name)
        except ValueError:
            raise ValueError("last target element not in source")

        old_source = source[: source_index + 1]

        if len(old_source) <= len(target):
            if target[-len(old_source) :] != old_source:
                raise ValueError(
                    "no clean match between end of target and beginning of source"
                )
        else:
            if target != source[source_index + 1 - len(target) : source_index + 1]:
                raise ValueError(
                    "no clean match between entire target and beginning of source"
                )

        return source[source_index + 1 :]

    @classmethod
    def _overlap_tail(
        cls,
        source: List[ComparisonItemType],
        target: List[ComparisonItemType],
    ) -> List[ComparisonItemType]:
        """
        Overlap must include first element of source.
        """

        source, target = cls._strip_none(source), cls._strip_none(target)

        if len(source) == 0 or len(target) == 0:
            return []

        if any((element is None for element in source)):
            raise ValueError("source is not consecutive")
        if any((element is None for element in target)):
            raise ValueError("target is not consecutive")

        source_names = {item.name for item in source}
        target_names = {item.name for item in target}

        if len(source_names) != len(source):
            raise ValueError("source contains duplicate entires")
        if len(target_names) != len(target):
            raise ValueError("target contains duplicate entires")

        overlap_tail = []
        for item in source:
            if item.name not in target_names:
                break
            overlap_tail.append(item)

        if len(overlap_tail) == 0:
            return overlap_tail

        target_index = target.index(overlap_tail[0])
        if overlap_tail != target[target_index : target_index + len(overlap_tail)]:
            raise ValueError("no clean match in overlap area")

        return overlap_tail

    @classmethod
    def _strip_none(
        cls, elements: List[ComparisonItemType]
    ) -> List[ComparisonItemType]:

        elements = cls._left_strip_none(elements)  # left strip
        elements.reverse()  # flip into reverse
        elements = cls._left_strip_none(elements)  # right strip
        elements.reverse()  # flip back

        return elements

    @staticmethod
    def _left_strip_none(
        elements: List[ComparisonItemType],
    ) -> List[ComparisonItemType]:

        return list(itertools.dropwhile(lambda element: element is None, elements))

    @staticmethod
    def _test_alternations(items: List[Union[SnapshotABC, None]]):

        alternations = 0
        state = False  # None

        for item in items:

            new_state = item is not None

            if new_state == state:
                continue

            alternations += 1
            state = new_state

            if alternations > 2:
                raise ValueError("gap in snapshot series")

    @staticmethod
    def _test_names(items: List[ComparisonItemABC]):

        for item in items:
            if item.a is not None and item.b is not None:
                if item.a.name != item.b.name:
                    raise ValueError("inconsistent snapshot names")

    @staticmethod
    def _find_name(snapshots: List[SnapshotABC], name: str) -> Union[int, None]:

        return next(
            (
                index
                for (index, snapshot) in enumerate(snapshots)
                if snapshot.name == name
            ),
            None,  # if nothing is found, return None
        )

    @staticmethod
    def _squash(snapshots: List[SnapshotABC]) -> List[SnapshotABC]:

        squashed, buffer = [], []

        for snapshot in snapshots:

            if snapshot.get("abgleich:type").value != "backup":
                buffer.append(snapshot)
            else:
                snapshot.intermediates.clear()
                snapshot.intermediates.extend(buffer)
                squashed.append(snapshot)
                buffer.clear()

        return squashed

    @classmethod
    def _merge_snapshots(
        cls,
        items_a: Generator[SnapshotABC, None, None],
        items_b: Generator[SnapshotABC, None, None],
        config: ConfigABC,
    ) -> List[ComparisonItemABC]:

        items_a, items_b = list(items_a), list(items_b)

        if config["compatibility/tagging"].value:
            items_a, items_b = cls._squash(items_a), cls._squash(items_b)

        if len(items_a) == 0 and len(items_b) == 0:
            return []

        assert len(set({item.name for item in items_a})) == len(items_a)  # unique names
        assert len(set({item.name for item in items_b})) == len(items_b)  # unique names

        if len(items_a) == 0:
            return [ComparisonItem(None, item) for item in items_b]
        if len(items_b) == 0:
            return [ComparisonItem(item, None) for item in items_a]

        start_b = cls._find_name(items_a, items_b[0].name)
        start_a = cls._find_name(items_b, items_a[0].name)

        assert start_a is not None or start_b is not None  # overlap

        if start_a is not None:  # fill prefix
            items_a = [None for _ in range(start_a)] + items_a
        if start_b is not None:  # fill prefix
            items_b = [None for _ in range(start_b)] + items_b
        if len(items_a) < len(items_b):  # fill suffix
            items_a = items_a + [None for _ in range(len(items_b) - len(items_a))]
        if len(items_b) < len(items_a):  # fill suffix
            items_b = items_b + [None for _ in range(len(items_a) - len(items_b))]

        assert len(items_a) == len(items_b)

        cls._test_alternations(items_a)
        cls._test_alternations(items_b)

        merged = [
            ComparisonItem(item_a, item_b) for item_a, item_b in zip(items_a, items_b)
        ]

        cls._test_names(merged)

        return merged

    @classmethod
    def from_datasets(
        cls,
        dataset_a: Union[DatasetABC, None],
        dataset_b: Union[DatasetABC, None],
        config: ConfigABC,
    ) -> ComparisonDatasetABC:

        assert dataset_a is not None or dataset_b is not None

        if (dataset_a is None) ^ (dataset_b is None):
            return cls(
                a=dataset_a,
                b=dataset_b,
                merged=ComparisonItem.list_from_singles(
                    getattr(dataset_a, "snapshots", None),
                    getattr(dataset_b, "snapshots", None),
                ),
            )

        assert dataset_a is not dataset_b
        assert dataset_a == dataset_b

        return cls(
            a=dataset_a,
            b=dataset_b,
            merged=cls._merge_snapshots(
                dataset_a.snapshots, dataset_b.snapshots, config
            ),
        )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/i18n.py: Translations

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import locale
import os
import warnings

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import FullLoader as Loader

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

from .debug import DEBUG, typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class _Lang(dict):
    """
    Mutable.
    """

    def __init__(self):

        super().__init__()
        self._lang = locale.getlocale()[0].split("_")[0]
        self._path = os.path.join(
            os.path.dirname(__file__), "..", "share", "translations.yaml"
        )
        self._load()

    def __call__(self, name: str) -> str:

        if len(name) == 0:
            if DEBUG:
                warnings.warn("empty text can not be translated", RuntimeWarning)
            return ""

        if int(os.environ.get("ABGLEICH_TRANSLATE", "0")) == 1:
            if name not in self.keys():
                warnings.warn(f"adding text for translation: {name:s}", RuntimeWarning)
                self._add_item(name)

        return self.get(name, {}).get(self._lang, name)

    def _add_item(self, name: str):

        self[name] = {}
        self._dump()

    def _load(self):

        self.clear()

        with open(self._path, "r") as f:
            self.update(yaml.load(f.read(), Loader=Loader))

    def _dump(self):

        with open(self._path, "w") as f:
            f.write(yaml.dump(self.copy(), Dumper=Dumper, allow_unicode=True, indent=4))


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# API
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

t = _Lang()



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/comparisonitem.py: ZFS comparison item

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Generator, List, Union

from .abc import ComparisonItemABC, DatasetABC, SnapshotABC
from .debug import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TYPING
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ComparisonGeneratorType = Union[
    Generator[DatasetABC, None, None],
    Generator[SnapshotABC, None, None],
    None,
]
ComparisonItemType = Union[
    DatasetABC,
    SnapshotABC,
    None,
]
ComparisonStrictItemType = Union[
    DatasetABC,
    SnapshotABC,
]

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class ComparisonItem(ComparisonItemABC):
    """
    Immutable.
    """

    def __init__(self, a: ComparisonItemType, b: ComparisonItemType):

        assert a is not None or b is not None
        if a is not None and b is not None:
            assert type(a) == type(b)

        self._a, self._b = a, b

    def get_item(self) -> ComparisonStrictItemType:

        if self._a is not None:
            return self._a
        return self._b

    @property
    def complete(self) -> bool:

        return self._a is not None and self._b is not None

    @property
    def a(self) -> ComparisonItemType:

        return self._a

    @property
    def b(self) -> ComparisonItemType:

        return self._b

    @classmethod
    def list_from_singles(
        cls,
        items_a: ComparisonGeneratorType,
        items_b: ComparisonGeneratorType,
    ) -> List[ComparisonItemABC]:

        assert (items_a is not None) ^ (items_b is not None)

        if items_a is None:
            return [cls(None, item) for item in items_b]
        return [cls(item, None) for item in items_a]



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/configfield.py: Handles configuration fields

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Callable, List, Optional, Union, Type

from .abc import ConfigFieldABC
from .debug import typechecked
from .i18n import t

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TYPING
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ConfigValueTypes = Union[List[str], str, int, float, bool, None]

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class ConfigField(ConfigFieldABC):
    """
    Mutable.
    """

    def __init__(
        self,
        name: str,
        description: str,
        type_: Type,
        validate: Optional[Callable] = None,
        default: ConfigValueTypes = None,
        import_: Optional[Callable] = None,
    ):

        self._name = name
        self._description = description
        self._type = type_
        self._default = default
        self._validate_func = validate if validate is not None else lambda v: True
        self._import = import_ if import_ is not None else lambda v: type_(v)

        if self._default is not None:
            if not self._validate(self._default):
                raise ValueError(f"invalid default value for {self._name}")

        self._value = None

    def __repr__(self) -> str:

        return (
            "<ConfigField "
            f'set="{"yes" if self.set else "no"}" '
            f'required="{str(self.required):s}" '
            f'type="{getattr(self._type, "__name__", str(type)):s}" '
            f'value="{str(self._value):s}" '
            f'default="{str(self._default):s}"'
            ">"
        )

    def copy(self) -> ConfigFieldABC:

        return type(self)(
            name=self._name,
            description=self._description,
            default=self._default,
            validate=self._validate_func,
            type_=self._type,
            import_=self._import,
        )

    def prompt(self):

        appendix = ""
        if self.required:
            appendix = f'({t("required"):s})'
        else:
            appendix = f'[{self._default}]'
        msg = t('Enter a valid value for') + f' "{self._name:s}" {appendix:s}. {self.description:s}'
        print(msg)

        while True:  # repeat until input value makes sense

            value = input('? ')

            if len(value) == 0:  # empty value, i.e. no value, given
                if self.required:  # value required, i.e. no default possible
                    print(t('Value required for') + f' "{self.name:s}". ' + t('Try again.'))
                    continue  # FAIL, try again
                else:
                    break  # OK, use default, leave
            else:  # some value at least one character long given
                try:
                    value = self._import(value)  # attempt to convert to type / import value
                except ValueError:
                    print(t('Can not be converted to type') + f' "{getattr(self._type, "__name__", str(type)):s}". ' + t('Try again.'))
                    continue  # FAIL, try again

            try:
                self.value = value  # attempt to set value
                break  # OK, value set, leave
            except ValueError:
                print(t('Not a valid value for') + f' "{self.name:s}". ' + t('Try again.'))
                continue  # FAIL, try again

    def _validate(self, value):

        return isinstance(value, self._type) and self._validate_func(value)

    @property
    def name(self) -> str:

        return self._name

    @property
    def description(self) -> str:

        return self._description

    @property
    def value(self) -> ConfigValueTypes:

        if self._value is not None:
            return self._value

        if self._default is None:
            raise ValueError(f"required value for {self._name} missing")

        return self._default

    @value.setter
    def value(self, value: ConfigValueTypes):

        if self._value is not None:
            raise ValueError(f"value for {self._name} has already been set")
        if value is None:
            return
        if not self._validate(value):
            raise ValueError(f"invalid value for {self._name}")

        self._value = value

    @property
    def valid(self) -> bool:

        if self._value is not None:
            return self._validate(self._value)

        # if self._default is not None
        return self._validate(self._default)

    @property
    def required(self) -> bool:

        return self._default is None

    @property
    def set(self) -> bool:

        return self._value is not None



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/config.py: Handles configuration data

    Copyright (C) 2019-2026 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/release_0.1/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Dict, TextIO, Union

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import FullLoader as Loader

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

from .abc import ConfigABC, ConfigFieldABC
from .configspec import CONFIGSPEC
from .debug import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class Config(ConfigABC):
    """
    Immutable.
    """

    def __init__(self, root: Union[str, None] = None, **kwargs: ConfigFieldABC):

        self._root = root
        self._fields = kwargs

    def __repr__(self):

        return "<Config>" if self._root is None else f'<Config root="{self._root:s}">'

    def __getitem__(self, key: str) -> ConfigFieldABC:

        return (
            self._fields[key]
            if self._root is None
            else self._fields[f"{self._root:s}/{key:s}"]
        )

    def group(self, root: str) -> ConfigABC:
        """
        Easy access to configuration groups (aka paths)
        """

        return type(self)(root=root, **self._fields)

    def to_fd(self, fd: TextIO):
        """
        Export configuration to handle on opened YAML file
        """

        fd.write(self.to_text())

    def to_text(self):
        """
        Export configuration to YAML string
        """

        config = {
            name: field.value for name, field in self._fields.items() if field.set
        }

        config = self._flat_to_tree(config)

        return yaml.dump(config, Dumper=Dumper, allow_unicode=True, indent=4)

    @classmethod
    def _tree_to_flat(cls, data: Dict, root: Union[str, None] = None) -> Dict:

        flat = {}

        for key, value in data.items():
            if not isinstance(key, str):
                raise TypeError("configuration key is no string", key)
            if root is not None:
                if len(root) > 0:
                    key = f"{root:s}/{key:s}"
            if isinstance(value, dict):
                flat.update(cls._tree_to_flat(data=value, root=key))
            else:
                flat[key] = value

        return flat

    @classmethod
    def _flat_to_tree(cls, data: Dict) -> Dict:

        tree = {}

        for key, value in data.items():

            if "/" not in key:
                tree[key] = value
                continue

            root, key = key.split("/", 1)
            if root not in tree.keys():
                tree[root] = {}
            tree[root][key] = value

        for key, value in tree.items():
            if not isinstance(value, dict):
                continue
            tree[key] = cls._flat_to_tree(value)

        return tree

    @classmethod
    def from_fd(cls, fd: TextIO) -> ConfigABC:
        """
        Import configuration from handle on opened YAML file
        """

        return cls.from_text(fd.read())

    @classmethod
    def from_text(cls, text: str) -> ConfigABC:
        """
        Import configuration from YAML string
        """

        raw = yaml.load(text, Loader=Loader)

        if not isinstance(raw, dict):
            raise TypeError("raw configuration is no dict", raw)

        config = {field.name: field.copy() for field in CONFIGSPEC}
        for key, value in cls._tree_to_flat(data=raw).items():
            if key not in config.keys():
                raise ValueError("unknown configuration key", key)
            config[key].value = value

        if any((not field.valid for field in config.values())):
            raise ValueError("configuration is not valid")

        return cls(**config)

    @classmethod
    def from_cli(cls):

        config = {field.name: field.copy() for field in CONFIGSPEC}

        source_keys = [key for key in config.keys() if key.startswith('source/')]
        target_keys = [key for key in config.keys() if key.startswith('target/')]
        other_keys = [key for key in config.keys() if key not in source_keys and key not in target_keys]

        source_keys.sort()
        target_keys.sort()
        other_keys.sort()

        for keys in (source_keys, target_keys, other_keys):
            for key in keys:
                config[key].prompt()

        return cls(**config)
